/*
* Copyright (c) 2003, University of Maryland 
* Copyright (c) 2003, Center for Satellite and Hybrid Communication Networks 
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions 
* are met:
*
*   * Redistributions of source code must retain the above copyright notice, 
*     this list of conditions,the authors and the following disclaimer.
*   * Redistributions in binary form must reproduce the above copyright notice,
*     this list of conditions,the authors and the following disclaimer in
*     the documentation and/or other materials provided with the distribution.
*   * Neither the name of the University of Maryland/Center for Satellite
*     and Hybrid Communication Networks nor the names of its 
*     contributors may be used to endorse or promote products derived from 
*     this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
* ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
* LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
* POSSIBILITY OF SUCH DAMAGE.
*
*/

/* Authors: 
*     Jonathan Polley (jpolley@isr.umd.edu),
*     Dionysus Blazakis (dblaze@isr.umd.edu),
*     Jonathan McGee (mcgee@isr.umd.edu),
*     Daniel Rusk (ruskd@isr.umd.edu),
*     Manish Karir (karir@isr.umd.edu)
*/

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include <gdk/gdkkeysyms.h>
#include <gtk/gtk.h>

#include "callbacks.h"
#include "interface.h"
#include "support.h"
#include "libelf.h"
#include "table.h"
#include "gtk_helper.h"
#include "chip_data.h"
#include "chip_data_struct.h"
#include "atemu.h"
#include "func_data.h"
#include "breakpoints.h"

int finished_loading_print = TRUE;
double vpane_ratio, hpane_ratio;

GtkItemFactory *item_factory;
static GtkItemFactoryEntry treeview_break_popup_item =
        { "/Clear Breakpoint", NULL, break_popup_clear_break,
             0, "<StockItem>", GTK_STOCK_REMOVE };

static GtkItemFactoryEntry treeview_watch_popup_item =
        { "/Clear Watchpoint", NULL, watch_popup_clear_watch,
            0, "<StockItem>", GTK_STOCK_REMOVE };

GtkTextTagTable *tagtable;

GtkWidget *notebook_chips;
    
GtkWidget *button_step;
GtkWidget *button_go;
GtkWidget *button_cstep;
GtkWidget *button_stop;
GtkWidget *button_restart;
GtkWidget *button_set_break;
GtkWidget *button_set_watch;

GtkWidget *textview_console;
GtkTextBuffer *textbuff_console;

GtkWidget *window1;

void create_chip_tab(GtkWidget *notebook_chips, int ntab){
    GtkWidget *label1;
    char tab_label[20];
    GtkWidget *scrolledwindow1;
    GtkWidget *notebook_data;
    GtkWidget *scrolledwindow6;
    GtkWidget *label6;
    GtkWidget *scrolledwindow7;
    GtkWidget *label7;
    GtkWidget *scrolledwindow13;
    GtkWidget *label16;
    GtkWidget *scrolledwindow14;
    GtkWidget *label17;
    GtkWidget *notebook_debugger;
    GtkWidget *scrolledwindow4;
    GtkWidget *label9;
    GtkWidget *scrolledwindow5;
    GtkWidget *label10;
    GtkWidget *scrolledwindow11;
    GtkWidget *label14;
    GtkWidget *scrolledwindow12;
    GtkWidget *label15;
    GtkWidget *notebook_code;
    GtkWidget *label18;
    GtkWidget *label19;
    GtkWidget *vbox1;
    GtkWidget *hbox1;
    GtkWidget *scrolledwindow15;
    GtkWidget *label20;
    GtkCellRenderer *renderer;
    GtkTreeViewColumn *column;
    chip_data *data = chip_array + ntab;
    int x, y, hpos, vpos;

    data->id = ntab;

    renderer = gtk_cell_renderer_text_new ();
    
    /* build gui elements.  Takes a lot of code */
    data->vpaned = gtk_vpaned_new ();
    gtk_widget_show (data->vpaned);
    gtk_container_add (GTK_CONTAINER (notebook_chips), data->vpaned);
    vpane_ratio = 0.55;
    g_signal_connect(G_OBJECT(data->vpaned), "notify::position", G_CALLBACK(vpane_change_position), NULL);

    data->hpaned = gtk_hpaned_new ();
    gtk_widget_show (data->hpaned);
    gtk_paned_pack1 (GTK_PANED (data->vpaned), data->hpaned, FALSE, TRUE);
    hpane_ratio = 0.65;
    g_signal_connect(G_OBJECT(data->hpaned), "notify::position", G_CALLBACK(hpane_change_position), NULL);

    notebook_code = gtk_notebook_new();
    gtk_widget_show(notebook_code);
    gtk_paned_pack1(GTK_PANED(data->hpaned), notebook_code, FALSE, TRUE);

    label18 = gtk_label_new("Disassembly");
    gtk_widget_show(label18);
    
    scrolledwindow1 = gtk_scrolled_window_new (NULL, NULL);
    gtk_widget_show (scrolledwindow1);
    gtk_notebook_append_page(GTK_NOTEBOOK(notebook_code), scrolledwindow1, label18);

    label19 = gtk_label_new("Source Code");
    gtk_widget_show(label19);

    vbox1 = gtk_vbox_new(FALSE, 0);
    gtk_widget_show(vbox1);
    gtk_notebook_append_page(GTK_NOTEBOOK(notebook_code), vbox1, label19);

    hbox1 = gtk_hbox_new(FALSE, 0);
    gtk_widget_show(hbox1);
    gtk_box_pack_start(GTK_BOX(vbox1), hbox1, FALSE, FALSE, 0);

    label20 = gtk_label_new("Source File:");
    gtk_widget_show(label20);
    gtk_box_pack_start(GTK_BOX(hbox1), label20, FALSE, FALSE, 0);

    data->cfile_select_menu = gtk_menu_new();
    
    data->cfile_select = gtk_option_menu_new();
    gtk_widget_show(data->cfile_select);
    gtk_box_pack_start(GTK_BOX(hbox1), data->cfile_select, TRUE, TRUE, 0);
    gtk_option_menu_set_menu(GTK_OPTION_MENU(data->cfile_select), data->cfile_select_menu);
    g_signal_connect( G_OBJECT( data->cfile_select ), "changed",
                G_CALLBACK( select_cfile ), data );

    scrolledwindow15 = gtk_scrolled_window_new (NULL, NULL);
    gtk_widget_show (scrolledwindow15);
    gtk_box_pack_start(GTK_BOX(vbox1), scrolledwindow15, TRUE, TRUE, 0);

    data->textview_source = gtk_text_view_new ();
    gtk_text_view_set_editable( GTK_TEXT_VIEW( data->textview_source ), FALSE );
    gtk_text_view_set_cursor_visible( GTK_TEXT_VIEW( data->textview_source ), FALSE );
    gtk_widget_show (data->textview_source);
    gtk_container_add (GTK_CONTAINER (scrolledwindow15), data->textview_source);
    g_signal_connect(G_OBJECT(data->textview_source), "populate-popup", G_CALLBACK(source_window_popup), NULL);
    g_signal_connect(G_OBJECT(data->textview_source), "button-press-event", G_CALLBACK(cursor_to_click), NULL);

    data->textview_code = gtk_text_view_new ();
    gtk_text_view_set_editable( GTK_TEXT_VIEW( data->textview_code ), FALSE );
    gtk_text_view_set_cursor_visible( GTK_TEXT_VIEW( data->textview_code ), FALSE );
    gtk_widget_show (data->textview_code);
    gtk_container_add (GTK_CONTAINER (scrolledwindow1), data->textview_code);
    g_signal_connect(G_OBJECT(data->textview_code), "populate-popup", G_CALLBACK(code_window_popup), NULL);
    g_signal_connect(G_OBJECT(data->textview_code), "button-press-event", G_CALLBACK(cursor_to_click), NULL);

    notebook_data = gtk_notebook_new ();
    gtk_notebook_set_scrollable(GTK_NOTEBOOK(notebook_data), TRUE);
    gtk_widget_show (notebook_data);
    gtk_paned_pack2 (GTK_PANED (data->hpaned), notebook_data, TRUE, TRUE);
    
    scrolledwindow6 = gtk_scrolled_window_new (NULL, NULL);
    gtk_widget_show (scrolledwindow6);
    gtk_container_add (GTK_CONTAINER (notebook_data), scrolledwindow6);

    data->treeview_regs = gtk_tree_view_new ();

    column = gtk_tree_view_column_new_with_attributes ("Register",
                                                   renderer,
                                                   "text", 0,
                                                   NULL);
    gtk_tree_view_append_column (GTK_TREE_VIEW (data->treeview_regs), column);

    column = gtk_tree_view_column_new_with_attributes ("Value (hex)",
                                                   renderer,
                                                   "text", 1,
                                                   NULL);
    gtk_tree_view_append_column (GTK_TREE_VIEW (data->treeview_regs), column);

    column = gtk_tree_view_column_new_with_attributes ("Value (dec)",
                                                   renderer,
                                                   "text", 2,
                                                   NULL);
    gtk_tree_view_append_column (GTK_TREE_VIEW (data->treeview_regs), column);

    gtk_widget_show (data->treeview_regs);
    gtk_container_add (GTK_CONTAINER (scrolledwindow6), data->treeview_regs);

    label6 = gtk_label_new (_("Registers"));
    gtk_widget_show (label6);
    gtk_notebook_set_tab_label (GTK_NOTEBOOK (notebook_data), data->treeview_regs->parent, label6);
    gtk_label_set_justify (GTK_LABEL (label6), GTK_JUSTIFY_LEFT);

    scrolledwindow7 = gtk_scrolled_window_new (NULL, NULL);
    gtk_widget_show (scrolledwindow7);
    gtk_container_add (GTK_CONTAINER (notebook_data), scrolledwindow7);

    data->treeview_stack = gtk_tree_view_new ();

    column = gtk_tree_view_column_new_with_attributes ("Call Stack - Top",
                                                   renderer,
                                                   "text", 0,
                                                   NULL);
    gtk_tree_view_append_column (GTK_TREE_VIEW (data->treeview_stack),
        column);

    gtk_widget_show (data->treeview_stack);
    gtk_container_add (GTK_CONTAINER (scrolledwindow7), data->treeview_stack);

    label7 = gtk_label_new (_("Stack"));
    gtk_widget_show (label7);
    gtk_notebook_set_tab_label (GTK_NOTEBOOK (notebook_data), data->treeview_stack->parent, label7);
    gtk_label_set_justify (GTK_LABEL (label7), GTK_JUSTIFY_LEFT);

    scrolledwindow13 = gtk_scrolled_window_new (NULL, NULL);
    gtk_widget_show (scrolledwindow13);
    gtk_container_add (GTK_CONTAINER (notebook_data), scrolledwindow13);

    data->treeview_locals = gtk_tree_view_new ();

    column = gtk_tree_view_column_new_with_attributes ("Local Name",
                                                   renderer,
                                                   "text", 0,
                                                   NULL);
    gtk_tree_view_append_column (GTK_TREE_VIEW (data->treeview_locals),
        column);

    column = gtk_tree_view_column_new_with_attributes ("Value",
                                                   renderer,
                                                   "text", 1,
                                                   NULL);
    gtk_tree_view_append_column (GTK_TREE_VIEW (data->treeview_locals),
        column);

    gtk_widget_show (data->treeview_locals);
    gtk_container_add (GTK_CONTAINER (scrolledwindow13), data->treeview_locals);

    g_signal_connect(G_OBJECT(data->treeview_locals), "button-press-event", G_CALLBACK(click_locals), NULL);

    label16 = gtk_label_new (_("Locals"));
    gtk_widget_show (label16);
    gtk_notebook_set_tab_label (GTK_NOTEBOOK (notebook_data), data->treeview_locals->parent, label16);
    gtk_label_set_justify (GTK_LABEL (label16), GTK_JUSTIFY_LEFT);

    scrolledwindow14 = gtk_scrolled_window_new (NULL, NULL);
    gtk_widget_show (scrolledwindow14);
    gtk_container_add (GTK_CONTAINER (notebook_data), scrolledwindow14);

    data->treeview_globals = gtk_tree_view_new ();

    column = gtk_tree_view_column_new_with_attributes ("Global Name",
                                                   renderer,
                                                   "text", 0,
                                                   NULL);
    gtk_tree_view_append_column (GTK_TREE_VIEW (data->treeview_globals),
        column);

    column = gtk_tree_view_column_new_with_attributes ("Address",
                                                   renderer,
                                                   "text", 1,
                                                   NULL);
    gtk_tree_view_append_column (GTK_TREE_VIEW (data->treeview_globals),
        column);

    column = gtk_tree_view_column_new_with_attributes ("Value (hex)",
                                                   renderer,
                                                   "text", 2,
                                                   NULL);
    gtk_tree_view_append_column (GTK_TREE_VIEW (data->treeview_globals),
        column);

    gtk_widget_show (data->treeview_globals);
    gtk_container_add (GTK_CONTAINER (scrolledwindow14), data->treeview_globals);

    label17 = gtk_label_new (_("Globals"));
    gtk_widget_show (label17);
    gtk_notebook_set_tab_label (GTK_NOTEBOOK (notebook_data), data->treeview_globals->parent, label17);
    gtk_label_set_justify (GTK_LABEL (label17), GTK_JUSTIFY_LEFT);

    g_signal_connect(G_OBJECT(data->treeview_globals), "button-press-event", G_CALLBACK(click_globals), NULL);

    notebook_debugger = gtk_notebook_new ();
    gtk_notebook_set_scrollable(GTK_NOTEBOOK(notebook_debugger), TRUE);
    gtk_widget_show (notebook_debugger);
    gtk_paned_pack2 (GTK_PANED (data->vpaned), notebook_debugger, TRUE, TRUE);

    scrolledwindow4 = gtk_scrolled_window_new (NULL, NULL);
    gtk_widget_show (scrolledwindow4);
    gtk_container_add (GTK_CONTAINER (notebook_debugger), scrolledwindow4);

    data->textview_events = gtk_text_view_new ();
    gtk_text_view_set_editable( GTK_TEXT_VIEW( data->textview_events ), FALSE );
    gtk_text_view_set_cursor_visible( GTK_TEXT_VIEW( data->textview_events ), FALSE );
    gtk_widget_show (data->textview_events);
    gtk_container_add (GTK_CONTAINER (scrolledwindow4), data->textview_events);

    label9 = gtk_label_new (_("Events"));
    gtk_widget_show (label9);
    gtk_notebook_set_tab_label (GTK_NOTEBOOK (notebook_debugger), data->textview_events->parent, label9);
    gtk_label_set_justify (GTK_LABEL (label9), GTK_JUSTIFY_LEFT);

    scrolledwindow5 = gtk_scrolled_window_new (NULL, NULL);
    gtk_widget_show (scrolledwindow5);
    gtk_container_add (GTK_CONTAINER (notebook_debugger), scrolledwindow5);

    data->textview_debug = gtk_text_view_new ();
    gtk_text_view_set_editable( GTK_TEXT_VIEW( data->textview_debug ), FALSE );
    gtk_text_view_set_cursor_visible( GTK_TEXT_VIEW( data->textview_debug ), FALSE );
    gtk_widget_show (data->textview_debug);
    gtk_container_add (GTK_CONTAINER (scrolledwindow5), data->textview_debug);

    label10 = gtk_label_new (_("Debug"));
    gtk_widget_show (label10);
    gtk_notebook_set_tab_label (GTK_NOTEBOOK (notebook_debugger), data->textview_debug->parent, label10);
    gtk_label_set_justify (GTK_LABEL (label10), GTK_JUSTIFY_LEFT);

    scrolledwindow11 = gtk_scrolled_window_new (NULL, NULL);
    gtk_widget_show (scrolledwindow11);
    gtk_container_add (GTK_CONTAINER (notebook_debugger), scrolledwindow11);

    /*popup menu for treeview break*/
    item_factory = gtk_item_factory_new( 
        GTK_TYPE_MENU, "<treeviewbreakpopup>", NULL );
    gtk_item_factory_create_item( item_factory, 
        &treeview_break_popup_item, NULL, 1 );
    data->treeview_break_popup =  GTK_WIDGET(gtk_item_factory_get_widget( item_factory, 
                    "<treeviewbreakpopup>"));

    data->treeview_break = gtk_tree_view_new ();
    
    column = gtk_tree_view_column_new_with_attributes ("Type",
                                                   renderer,
                                                   "text", 0,
                                                   NULL);
    gtk_tree_view_append_column (GTK_TREE_VIEW (data->treeview_break), column);

    column = gtk_tree_view_column_new_with_attributes ("Location",
                                                   renderer,
                                                   "text", 1,
                                                   NULL);
    gtk_tree_view_append_column (GTK_TREE_VIEW (data->treeview_break), column);
    
    g_signal_connect(G_OBJECT(data->treeview_break), "button-press-event", G_CALLBACK(break_window_popup), NULL);
    gtk_widget_show (data->treeview_break);
    gtk_container_add (GTK_CONTAINER (scrolledwindow11), data->treeview_break);

    label14 = gtk_label_new (_("Breakpoints"));
    gtk_widget_show (label14);
    gtk_notebook_set_tab_label (GTK_NOTEBOOK (notebook_debugger), data->treeview_break->parent, label14);
    gtk_label_set_justify (GTK_LABEL (label14), GTK_JUSTIFY_LEFT);

    scrolledwindow12 = gtk_scrolled_window_new (NULL, NULL);
    gtk_widget_show (scrolledwindow12);
    gtk_container_add (GTK_CONTAINER (notebook_debugger), scrolledwindow12);

    /*popup menu for treeview break*/
    item_factory = gtk_item_factory_new( 
        GTK_TYPE_MENU, "<treeviewwatchpopup>", NULL );
    gtk_item_factory_create_item( item_factory, 
        &treeview_watch_popup_item, NULL, 1 );
    data->treeview_watch_popup =  GTK_WIDGET(gtk_item_factory_get_widget( item_factory, 
                    "<treeviewwatchpopup>"));
    
    data->treeview_watch = gtk_tree_view_new ();
    
    column = gtk_tree_view_column_new_with_attributes ("Location",
                                                   renderer,
                                                   "text", 0,
                                                   NULL);
    gtk_tree_view_append_column (GTK_TREE_VIEW (data->treeview_watch), column);
    
    g_signal_connect(G_OBJECT(data->treeview_watch), "button-press-event", G_CALLBACK(watch_window_popup), NULL);
    gtk_widget_show (data->treeview_watch);
    gtk_container_add (GTK_CONTAINER (scrolledwindow12), data->treeview_watch);

    label15 = gtk_label_new (_("Watchpoints"));
    gtk_widget_show (label15);
    gtk_notebook_set_tab_label (GTK_NOTEBOOK (notebook_debugger), data->treeview_watch->parent, label15);
    gtk_label_set_justify (GTK_LABEL (label15), GTK_JUSTIFY_LEFT);

    snprintf(tab_label, sizeof(tab_label), "%s %d", _("Chip"), ntab);
    tab_label[sizeof(tab_label)-1] = '\0';
    label1 = gtk_label_new (tab_label);
    gtk_widget_show (label1);
    gtk_notebook_set_tab_label (GTK_NOTEBOOK (notebook_chips), gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook_chips), ntab + 1), label1);
    gtk_label_set_justify (GTK_LABEL (label1), GTK_JUSTIFY_LEFT);

    /* set text buffers and tree models */    
    gtk_text_view_set_buffer( GTK_TEXT_VIEW( data->textview_code ), data->code );
    gtk_tree_view_set_model( GTK_TREE_VIEW( data->treeview_regs ), GTK_TREE_MODEL( data->regs ) );
    gtk_tree_view_set_model( GTK_TREE_VIEW( data->treeview_stack ), GTK_TREE_MODEL( data->stack ) );
    gtk_tree_view_set_model( GTK_TREE_VIEW( data->treeview_locals ), GTK_TREE_MODEL( data->locals ) );
    gtk_tree_view_set_model( GTK_TREE_VIEW( data->treeview_globals ), GTK_TREE_MODEL( data->globals ) );
    gtk_text_view_set_buffer( GTK_TEXT_VIEW( data->textview_events ), data->events );
    gtk_text_view_set_buffer( GTK_TEXT_VIEW( data->textview_debug ), data->debug );
    gtk_tree_view_set_model( GTK_TREE_VIEW( data->treeview_break ), GTK_TREE_MODEL( data->breakp ) );
    gtk_tree_view_set_model( GTK_TREE_VIEW( data->treeview_watch ), GTK_TREE_MODEL( data->watchp ) );

    /* set initial pane positions */
    gtk_window_get_size(GTK_WINDOW(window1), &x, &y);
    hpos = (int)(hpane_ratio * ((double)x));
    vpos = (int)(vpane_ratio * ((double)y));

    gtk_paned_set_position(GTK_PANED (data->hpaned), hpos);
    gtk_paned_set_position(GTK_PANED (data->vpaned), vpos);
}

/* this assumes each buffer contains exactly ONE 
   line of output... */  
void process_buffer( char *buff )
{
    int chipid, left, num_chips;
    uint64_t cycles;
    char label[64];
    char fname[128];
    char *data;
    GtkTextMark *mark;
    static int currchip=-1;
    long int addr = 0;
    int id;
    int value;

    /* for asm code loading */
    static long int currline = 0;
    static struct func_data *oldfunc = NULL;

    if(strcmp(buff, "[atemu]('?' for help) \n") == 0){
        pipe_write_safe = 1;
        if( finished_loading_print ){ 
            mark = text_buffer_printf( textbuff_console, "...Finished Loading!\n" );
            gtk_text_view_scroll_to_mark(GTK_TEXT_VIEW(textview_console), mark, 0.0, FALSE, 0.0, 0.0);
            finished_loading_print = FALSE;
        }
        /* need the atemu_going state flag to get set right */
        atemu_going = 0;
    }else if (strncmp(buff, "Loaded file ", sizeof("Loaded file ") - 1 ) == 0 ) {
        sscanf(buff, "Loaded file %128s for chip %d.", fname, &chipid );
        parse_elf( fname, chipid );
    } else if (strncmp(buff, "Number of ", sizeof("Number of ") - 1 ) == 0 ) {
        sscanf(buff, "Number of chips: %d", &num_chips );
        create_all_chips( num_chips );
    } else if(strncmp(buff, "Current chip ", sizeof("Current chip ")-1)==0){
        sscanf(buff, "Current chip is now %d\n", &currchip);
        currline = 0;
        oldfunc = NULL;
    } else if(strncmp(buff, "Call stack ", sizeof("Call stack ") -1) == 0 ) {
        sscanf(buff, "Call stack %d begin:\n", &currchip );
        gtk_list_store_clear( chip_array[currchip].stack );
    } else if(strncmp(buff, "\taddr: ", sizeof("\taddr: ") -1) == 0 ) {
        sscanf(buff, "\taddr: %lx\n", &addr );
        add_stack_item( currchip, addr );
    } else if(strncmp(buff, "notify ", sizeof("notify ") - 1) == 0) {
        sscanf(buff, "notify addr %lx id %d = %x, chip %d\n", &addr, &id,
            &value, &chipid );
        update_global( chipid, id, addr, value );
    }else if(strncmp(buff, "A0x", 3)==0 && currchip > -1) {
        struct func_data *func;
        
        sscanf(buff, "A%lx: ", &addr );

        /* see if we are in a different function than before */
        func = address_to_func_data(currchip, addr);
        if(func != oldfunc){
            GtkTextIter start, end;
            /* print the function name, and make it pretty */
            if(func != NULL){
                mark = text_buffer_printf(chip_array[currchip].code, "\n%s\n", func->name);
            }else{
                mark = text_buffer_printf(chip_array[currchip].code, "\n<unknown>\n");
            }
            gtk_text_buffer_get_iter_at_mark(chip_array[currchip].code, &start, mark);
            gtk_text_buffer_get_iter_at_line(chip_array[currchip].code, &end, gtk_text_iter_get_line(&start) - 1);
            gtk_text_buffer_apply_tag_by_name(chip_array[currchip].code, "function", &start, &end);

            currline+=2;
            oldfunc = func;
        }
        
        add_addr_line_mapping( currchip, addr, currline );
        text_buffer_printf(chip_array[currchip].code, "%s", buff);
        currline++;
    }else{
        if ( sscanf( buff, "%d %lld %s %n", &chipid, &cycles, label, &left ) )
        {
            data = buff + left;

            if ( strcmp( "event", label ) == 0 ){
                mark = text_buffer_printf( chip_array[chipid].events, "%16lld: %s", cycles, data);
                gtk_text_view_scroll_to_mark(GTK_TEXT_VIEW(chip_array[chipid].textview_events), mark, 0.0, FALSE, 0.0, 0.0);
                
                mark = text_buffer_printf( textbuff_console, "%s", buff );
                gtk_text_view_scroll_to_mark(GTK_TEXT_VIEW(textview_console), mark, 0.0, FALSE, 0.0, 0.0);
            } else if ( strcmp( "debug", label ) == 0 ) {
                // c stepping too slow if we print every breakpoint
                if(strncmp("breakpoint", data, 10) != 0){
                    mark = text_buffer_printf( chip_array[chipid].debug, "%16lld: %s", cycles, data);
                    gtk_text_view_scroll_to_mark(GTK_TEXT_VIEW(chip_array[chipid].textview_debug), mark, 0.0, FALSE, 0.0, 0.0);
                    
                    mark = text_buffer_printf( textbuff_console, "%s", buff );
                    gtk_text_view_scroll_to_mark(GTK_TEXT_VIEW(textview_console), mark, 0.0, FALSE, 0.0, 0.0);
                } else{ //don't filter valid breakpoints;
                    long int test_addr;
                    sscanf( data + 10, "%lx", &test_addr );
                    if( is_asm_breakpoint( chipid, test_addr ) || 
                        is_source_breakpoint( chipid, test_addr ) ){
                        mark = text_buffer_printf( chip_array[chipid].debug, "%16lld: %s", cycles, data);
                        gtk_text_view_scroll_to_mark(GTK_TEXT_VIEW(chip_array[chipid].textview_debug), mark, 0.0, FALSE, 0.0, 0.0);

                        mark = text_buffer_printf( textbuff_console, "%s", buff );
                        gtk_text_view_scroll_to_mark(GTK_TEXT_VIEW(textview_console), mark, 0.0, FALSE, 0.0, 0.0);
                    } 
                }  
            } else if ( strcmp( "data", label ) == 0 ) {
                int c;
                long data_addr;
                unsigned int dumpdata[16];
                uint8_t dumpdata_fixed[16];

                sscanf( data, "%lx: %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x %x", &data_addr,
                    &dumpdata[0], &dumpdata[1], 
                    &dumpdata[2], &dumpdata[3], 
                    &dumpdata[4], &dumpdata[5], 
                    &dumpdata[6], &dumpdata[7], 
                    &dumpdata[8], &dumpdata[9], 
                    &dumpdata[10], &dumpdata[11], 
                    &dumpdata[12], &dumpdata[13], 
                    &dumpdata[14], &dumpdata[15] );

                for( c = 0; c < 16; c++ )
                    dumpdata_fixed[c] = (uint8_t) dumpdata[c];

                if ( ( data_addr == 0x0000 ) || ( data_addr == 0x0010 ) ||
                     ( data_addr == 0x0050 ) ) {
                    refresh_registers( chipid, data_addr, dumpdata_fixed );
                }

                new_locals( chipid, data_addr, dumpdata_fixed );
                
                /* Useful for debugging 
                 * mark = text_buffer_printf( textbuff_console, "%s", buff );
                 * gtk_text_view_scroll_to_mark(GTK_TEXT_VIEW(textview_console), mark, 0.0, FALSE, 0.0, 0.0);
                 */
            }else if (strcmp("opcode", label) == 0){
                /* need to highlight the line in the asm code view */
                /* also in the C view */
                int line;
                long addr;
                char *cfile=NULL;
                GtkTextIter start, end;

                /* look up asm line */
                sscanf(data, "A%lx", &addr);

                /* update register tab: PC */
                set_register( chipid, REGS_TAB_PC, addr, 6, 7 );

                gtk_widget_set_sensitive( button_step, TRUE );
                gtk_widget_set_sensitive( button_cstep, TRUE );
                gtk_widget_set_sensitive( button_go, TRUE );
                gtk_widget_set_sensitive( button_stop, FALSE );

                line = address_to_text_line(chipid, addr);

                if(line >= 0){
                    /* remove any old highlighting */
                    gtk_text_buffer_get_start_iter(chip_array[chipid].code, &start);
                    gtk_text_buffer_get_end_iter(chip_array[chipid].code, &end);
                    gtk_text_buffer_remove_tag_by_name(chip_array[chipid].code, "current line", &start, &end);

                    /* and mark current line */
                    gtk_text_buffer_get_iter_at_line(chip_array[chipid].code, &start, line);
                    gtk_text_buffer_get_iter_at_line(chip_array[chipid].code, &end, line+1);
                    gtk_text_buffer_apply_tag_by_name(chip_array[chipid].code, "current line", &start, &end);

                    /* scroll to buffer */
                    gtk_text_view_scroll_to_iter(GTK_TEXT_VIEW(chip_array[chipid].textview_code), &start, 0.1, FALSE, 0.0, 0.0);
                }

                /* C line stuff */
                line = address_to_c_line(chipid, addr, &cfile);
                if(line >= 1){
                    GtkTextBuffer *buff;
                    int index;

                    buff = cfile_to_textbuffer(chipid, cfile);
                    index = cfile_to_cfile_index(chipid, cfile);
                    if(buff != NULL){
                        /* remove any old highlighting */
                        gtk_text_buffer_get_start_iter(buff, &start);
                        gtk_text_buffer_get_end_iter(buff, &end);
                        gtk_text_buffer_remove_tag_by_name(buff, "current line", &start, &end);

                        /* and mark current line */
                        /* text file line 1 is text buffer line 0 */
                        gtk_text_buffer_get_iter_at_line(buff, &start, line - 1);
                        gtk_text_buffer_get_iter_at_line(buff, &end, line);
                        gtk_text_buffer_apply_tag_by_name(buff, "current line", &start, &end);

                        /* load buffer and scroll to iter */
                        if( gtk_text_view_get_buffer(GTK_TEXT_VIEW(chip_array[chipid].textview_source)) != buff )
                            gtk_option_menu_set_history(GTK_OPTION_MENU(chip_array[chipid].cfile_select), index);

                        gtk_text_view_scroll_to_iter(GTK_TEXT_VIEW(chip_array[chipid].textview_source), &start, 0.1, FALSE, 0.0, 0.0);
                    }
                }

                atemu_dump_regs( chipid );

                update_locals( chipid, addr );
                
                mark = text_buffer_printf( textbuff_console, "%s", buff );
                gtk_text_view_scroll_to_mark(GTK_TEXT_VIEW(textview_console), mark, 0.0, FALSE, 0.0, 0.0);
            }
        }
    }
}

GtkWidget* create_window1 ()
{
    GtkWidget *vbox1;
    GtkWidget *menubar1;
    GtkWidget *menuitem1;
    GtkWidget *menu1;
    GtkWidget *new1;
    GtkWidget *open1;
    GtkWidget *save1;
    GtkWidget *save_as1;
    GtkWidget *separatormenuitem1;
    GtkWidget *quit1;
    GtkWidget *menuitem2;
    GtkWidget *menu2;
    GtkWidget *cut1;
    GtkWidget *copy1;
    GtkWidget *paste1;
    GtkWidget *delete1;
    GtkWidget *menuitem3;
    GtkWidget *menu3;
    GtkWidget *menuitem4;
    GtkWidget *menu4;
    GtkWidget *about1;
    GtkWidget *toolbar1;
    GtkWidget *tmp_toolbar_icon;
    GtkWidget *scrolledwindow3;
    GtkWidget *label8;
    GtkAccelGroup *accel_group;

    accel_group = gtk_accel_group_new ();

    window1 = gtk_window_new (GTK_WINDOW_TOPLEVEL);
    gtk_window_set_title (GTK_WINDOW (window1), _("Visual AVR Debugger & Simulator"));
    gtk_widget_set_size_request( window1, 750, 550 );
    g_signal_connect (G_OBJECT (window1), "destroy", G_CALLBACK (on_quit1_activate), NULL);
    g_signal_connect (G_OBJECT (window1), "delete_event", G_CALLBACK (on_quit1_activate), NULL);
    g_signal_connect (G_OBJECT (window1), "configure_event", G_CALLBACK (on_window1_configure), NULL);

    vbox1 = gtk_vbox_new (FALSE, 0);
    gtk_widget_show (vbox1);
    gtk_container_add (GTK_CONTAINER (window1), vbox1);

    menubar1 = gtk_menu_bar_new ();
    gtk_widget_show (menubar1);
    gtk_box_pack_start (GTK_BOX (vbox1), menubar1, FALSE, FALSE, 0);

    menuitem1 = gtk_menu_item_new_with_mnemonic (_("_File"));
    gtk_widget_show (menuitem1);
    gtk_container_add (GTK_CONTAINER (menubar1), menuitem1);

    menu1 = gtk_menu_new ();
    gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitem1), menu1);

    new1 = gtk_image_menu_item_new_from_stock ("gtk-new", accel_group);
    gtk_widget_show (new1);
    gtk_container_add (GTK_CONTAINER (menu1), new1);

    open1 = gtk_image_menu_item_new_from_stock ("gtk-open", accel_group);
    gtk_widget_show (open1);
    gtk_container_add (GTK_CONTAINER (menu1), open1);

    save1 = gtk_image_menu_item_new_from_stock ("gtk-save", accel_group);
    gtk_widget_show (save1);
    gtk_container_add (GTK_CONTAINER (menu1), save1);

    save_as1 = gtk_image_menu_item_new_from_stock ("gtk-save-as", accel_group);
    gtk_widget_show (save_as1);
    gtk_container_add (GTK_CONTAINER (menu1), save_as1);

    separatormenuitem1 = gtk_separator_menu_item_new ();
    gtk_widget_show (separatormenuitem1);
    gtk_container_add (GTK_CONTAINER (menu1), separatormenuitem1);
    gtk_widget_set_sensitive (separatormenuitem1, FALSE);

    quit1 = gtk_image_menu_item_new_from_stock ("gtk-quit", accel_group);
    gtk_widget_show (quit1);
    gtk_container_add (GTK_CONTAINER (menu1), quit1);

    menuitem2 = gtk_menu_item_new_with_mnemonic (_("_Edit"));
    gtk_widget_show (menuitem2);
    gtk_container_add (GTK_CONTAINER (menubar1), menuitem2);

    menu2 = gtk_menu_new ();
    gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitem2), menu2);

    cut1 = gtk_image_menu_item_new_from_stock ("gtk-cut", accel_group);
    gtk_widget_show (cut1);
    gtk_container_add (GTK_CONTAINER (menu2), cut1);

    copy1 = gtk_image_menu_item_new_from_stock ("gtk-copy", accel_group);
    gtk_widget_show (copy1);
    gtk_container_add (GTK_CONTAINER (menu2), copy1);

    paste1 = gtk_image_menu_item_new_from_stock ("gtk-paste", accel_group);
    gtk_widget_show (paste1);
    gtk_container_add (GTK_CONTAINER (menu2), paste1);

    delete1 = gtk_image_menu_item_new_from_stock ("gtk-delete", accel_group);
    gtk_widget_show (delete1);
    gtk_container_add (GTK_CONTAINER (menu2), delete1);

    menuitem3 = gtk_menu_item_new_with_mnemonic (_("_View"));
    gtk_widget_show (menuitem3);
    gtk_container_add (GTK_CONTAINER (menubar1), menuitem3);

    menu3 = gtk_menu_new ();
    gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitem3), menu3);

    menuitem4 = gtk_menu_item_new_with_mnemonic (_("_Help"));
    gtk_widget_show (menuitem4);
    gtk_container_add (GTK_CONTAINER (menubar1), menuitem4);

    menu4 = gtk_menu_new ();
    gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitem4), menu4);

    about1 = gtk_menu_item_new_with_mnemonic (_("_About"));
    gtk_widget_show (about1);
    gtk_container_add (GTK_CONTAINER (menu4), about1);

    toolbar1 = gtk_toolbar_new ();
    gtk_widget_show (toolbar1);
    gtk_box_pack_start (GTK_BOX (vbox1), toolbar1, FALSE, FALSE, 0);
    gtk_toolbar_set_style (GTK_TOOLBAR (toolbar1), GTK_TOOLBAR_BOTH);

    tmp_toolbar_icon = gtk_image_new_from_stock ("gtk-go-forward", gtk_toolbar_get_icon_size (GTK_TOOLBAR (toolbar1)));
    button_step = gtk_toolbar_append_element (GTK_TOOLBAR (toolbar1),
            GTK_TOOLBAR_CHILD_BUTTON,
            NULL,
            _("Step"),
            NULL, NULL,
            tmp_toolbar_icon, NULL, NULL);
    gtk_label_set_use_underline (GTK_LABEL (((GtkToolbarChild*) (g_list_last (GTK_TOOLBAR (toolbar1)->children)->data))->label), TRUE);
    g_signal_connect( G_OBJECT( button_step ), "clicked", G_CALLBACK( button_step_clicked ), NULL );
    gtk_widget_set_sensitive( button_step, FALSE );
    gtk_widget_show (button_step);

    tmp_toolbar_icon = gtk_image_new_from_stock ("gtk-jump-to", gtk_toolbar_get_icon_size (GTK_TOOLBAR (toolbar1)));
    button_cstep = gtk_toolbar_append_element (GTK_TOOLBAR (toolbar1),
            GTK_TOOLBAR_CHILD_BUTTON,
            NULL,
            _("C Step"),
            NULL, NULL,
            tmp_toolbar_icon, NULL, NULL);
    gtk_label_set_use_underline (GTK_LABEL (((GtkToolbarChild*) (g_list_last (GTK_TOOLBAR (toolbar1)->children)->data))->label), TRUE);
    g_signal_connect( G_OBJECT( button_cstep ), "clicked", G_CALLBACK( button_cstep_clicked ), NULL );
    gtk_widget_set_sensitive( button_cstep, FALSE );
    gtk_widget_show (button_cstep);

    tmp_toolbar_icon = gtk_image_new_from_stock ("gtk-execute", gtk_toolbar_get_icon_size (GTK_TOOLBAR (toolbar1)));
    button_go = gtk_toolbar_append_element (GTK_TOOLBAR (toolbar1),
            GTK_TOOLBAR_CHILD_BUTTON,
            NULL,
            _("Go"),
            NULL, NULL,
            tmp_toolbar_icon, NULL, NULL);
    gtk_label_set_use_underline (GTK_LABEL (((GtkToolbarChild*) (g_list_last (GTK_TOOLBAR (toolbar1)->children)->data))->label), TRUE);
    g_signal_connect( G_OBJECT( button_go ), "clicked", G_CALLBACK( button_go_clicked ), NULL );
    gtk_widget_set_sensitive( button_go, FALSE );
    gtk_widget_show (button_go);

    tmp_toolbar_icon = gtk_image_new_from_stock ("gtk-stop", gtk_toolbar_get_icon_size (GTK_TOOLBAR (toolbar1)));
    button_stop = gtk_toolbar_append_element (GTK_TOOLBAR (toolbar1),
            GTK_TOOLBAR_CHILD_BUTTON,
            NULL,
            _("Stop"),
            NULL, NULL,
            tmp_toolbar_icon, NULL, NULL);
    gtk_label_set_use_underline (GTK_LABEL (((GtkToolbarChild*) (g_list_last (GTK_TOOLBAR (toolbar1)->children)->data))->label), TRUE);
    g_signal_connect( G_OBJECT( button_stop ), "clicked", G_CALLBACK( button_stop_clicked ), NULL );
    gtk_widget_set_sensitive( button_stop, FALSE );
    gtk_widget_show (button_stop);

    tmp_toolbar_icon = gtk_image_new_from_stock ("gtk-refresh", gtk_toolbar_get_icon_size (GTK_TOOLBAR (toolbar1)));
    button_restart = gtk_toolbar_append_element (GTK_TOOLBAR (toolbar1),
            GTK_TOOLBAR_CHILD_BUTTON,
            NULL,
            _("Restart"),
            NULL, NULL,
            tmp_toolbar_icon, NULL, NULL);
    gtk_label_set_use_underline (GTK_LABEL (((GtkToolbarChild*) (g_list_last (GTK_TOOLBAR (toolbar1)->children)->data))->label), TRUE);
    g_signal_connect( G_OBJECT( button_restart ), "clicked", G_CALLBACK( button_restart_clicked ), NULL );
    gtk_widget_set_sensitive( button_restart, FALSE );
    gtk_widget_show (button_restart);

    tmp_toolbar_icon = gtk_image_new_from_stock ("gtk-add", gtk_toolbar_get_icon_size (GTK_TOOLBAR (toolbar1)));
    button_set_break = gtk_toolbar_append_element (GTK_TOOLBAR (toolbar1),
            GTK_TOOLBAR_CHILD_BUTTON,
            NULL,
            _("Set Breakpoint"),
            NULL, NULL,
            tmp_toolbar_icon, NULL, NULL);
    gtk_label_set_use_underline (GTK_LABEL (((GtkToolbarChild*) (g_list_last (GTK_TOOLBAR (toolbar1)->children)->data))->label), TRUE);
    g_signal_connect( G_OBJECT( button_set_break ), "clicked", G_CALLBACK( button_set_break_clicked ), NULL );
    gtk_widget_set_sensitive( button_set_break, FALSE );
    gtk_widget_show (button_set_break);

    tmp_toolbar_icon = gtk_image_new_from_stock ("gtk-add", gtk_toolbar_get_icon_size (GTK_TOOLBAR (toolbar1)));
    button_set_watch = gtk_toolbar_append_element (GTK_TOOLBAR (toolbar1),
            GTK_TOOLBAR_CHILD_BUTTON,
            NULL,
            _("Set Watchpoint"),
            NULL, NULL,
            tmp_toolbar_icon, NULL, NULL);
    gtk_label_set_use_underline (GTK_LABEL (((GtkToolbarChild*) (g_list_last (GTK_TOOLBAR (toolbar1)->children)->data))->label), TRUE);
    g_signal_connect( G_OBJECT( button_set_watch ), "clicked", G_CALLBACK( button_set_watch_clicked ), NULL );
    gtk_widget_set_sensitive( button_set_watch, FALSE );
    gtk_widget_show (button_set_watch);

    notebook_chips = gtk_notebook_new ();
    gtk_notebook_set_scrollable(GTK_NOTEBOOK(notebook_chips), TRUE);
    gtk_widget_show (notebook_chips);
    g_signal_connect(G_OBJECT(notebook_chips), "switch-page", G_CALLBACK(notebook_chips_select), NULL);
    gtk_box_pack_start (GTK_BOX (vbox1), notebook_chips, TRUE, TRUE, 0);

    scrolledwindow3 = gtk_scrolled_window_new (NULL, NULL);
    gtk_widget_show (scrolledwindow3);
    gtk_container_add (GTK_CONTAINER (notebook_chips), scrolledwindow3);

    textview_console = gtk_text_view_new ();
    gtk_text_view_set_editable( GTK_TEXT_VIEW( textview_console ), FALSE );
    gtk_text_view_set_cursor_visible( GTK_TEXT_VIEW( textview_console ), FALSE );
    textbuff_console = gtk_text_view_get_buffer( GTK_TEXT_VIEW( textview_console ) );
    gtk_widget_show (textview_console);

    tagtable = gtk_text_buffer_get_tag_table(textbuff_console);

    /* create text style tags */
    gtk_text_buffer_create_tag(textbuff_console, "current line",
        "background-gdk", &(textview_console->style->bg[GTK_STATE_SELECTED]),
        "foreground-gdk", &(textview_console->style->fg[GTK_STATE_SELECTED]),
        NULL);
    gtk_text_buffer_create_tag(textbuff_console, "breakpoint", 
        "foreground", "red", 
        NULL);
    gtk_text_buffer_create_tag(textbuff_console, "function", 
        "foreground", "dimgrey", 
        "style", PANGO_STYLE_ITALIC,
        NULL);
    gtk_text_buffer_create_tag(textbuff_console, "monospace", 
        "family", "courier", 
        NULL);

    gtk_container_add (GTK_CONTAINER (scrolledwindow3), textview_console);

    label8 = gtk_label_new (_("Console"));
    gtk_widget_show (label8);
    gtk_notebook_set_tab_label (GTK_NOTEBOOK (notebook_chips), gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook_chips), 0), label8);
    gtk_label_set_justify (GTK_LABEL (label8), GTK_JUSTIFY_LEFT);

    g_signal_connect ((gpointer) new1, "activate",
            G_CALLBACK (on_new1_activate),
            NULL);
    g_signal_connect ((gpointer) open1, "activate",
            G_CALLBACK (on_open1_activate),
            NULL);
    g_signal_connect ((gpointer) save1, "activate",
            G_CALLBACK (on_save1_activate),
            NULL);
    g_signal_connect ((gpointer) save_as1, "activate",
            G_CALLBACK (on_save_as1_activate),
            NULL);
    g_signal_connect ((gpointer) quit1, "activate",
            G_CALLBACK (on_quit1_activate),
            NULL);
    g_signal_connect ((gpointer) cut1, "activate",
            G_CALLBACK (on_cut1_activate),
            NULL);
    g_signal_connect ((gpointer) copy1, "activate",
            G_CALLBACK (on_copy1_activate),
            NULL);
    g_signal_connect ((gpointer) paste1, "activate",
            G_CALLBACK (on_paste1_activate),
            NULL);
    g_signal_connect ((gpointer) delete1, "activate",
            G_CALLBACK (on_delete1_activate),
            NULL);
    g_signal_connect ((gpointer) about1, "activate",
            G_CALLBACK (on_about1_activate),
            NULL);

    gtk_window_add_accel_group (GTK_WINDOW (window1), accel_group);

    return window1;
}
